@@ -40,13 +40,18 @@
 ## -- 2023-01-14  1.9.1     SY       Add class Label
 ## -- 2023-01-31  1.9.2     SY       Renaming class Label to PersonalisedStamp
 ## -- 2023-02-22  2.0.0     DA       Class Saveable: new custom method _save()
+## -- 2023-03-27  2.1.0     DA       Refactoring persistence:
+## --                                - new class Id
+## --                                - renamed class LoadSave to Persistent
+## --                                - merged classes Load, Save into Persistent
+## --                                - logging
 ## -------------------------------------------------------------------------------------------------
 
 """
-Ver. 2.0.0 (2023-02-22)
+Ver. 2.1.0 (2023-03-27)
 
 This module provides various classes with elementry functionalities for reuse in higher level classes. 
-For example: logging, load/save, timer...
+For example: logging, persistence, timer...
 """
 
 
@@ -54,122 +59,59 @@ from datetime import datetime, timedelta
 from time import sleep
 import dill as pkl
 import os
+import sys
 import uuid
 from mlpro.bf.exceptions import *
 
 
+# Global dictionary to store paths of pickle files during runtime
+g_persistence_file_paths = {}
+
+
 
 
 ## -------------------------------------------------------------------------------------------------
 ## -------------------------------------------------------------------------------------------------
-class Loadable:
+class Id:
     """
-    This abstract class adds the ability to be loadable to inherited classes. 
+    Property class that inherits a unique id and related get/set-methods to a child class.
+
+    Parameters
+    ----------
+    p_id
+        Optional external id
+
+    Attributes
+    _id
+        Unique id of the object.
+
     """
 
 ## -------------------------------------------------------------------------------------------------
-    @staticmethod
-    def load(p_path, p_filename):
-        """
-        Loads content from the given path and file name. If file does not exist, it returns None.
-
-        Parameters:
-            p_path          Path that contains the file 
-            p_filename      File name
-
-        Returns: 
-            A loaded object, if file content was loaded successfully. None otherwise.
-        """
-
-        if not os.path.exists(p_path + os.sep + p_filename): return None
-
-        return pkl.load(open(p_path + os.sep + p_filename, 'rb'))
-
-
-
-
-
-## -------------------------------------------------------------------------------------------------
-## -------------------------------------------------------------------------------------------------
-class Saveable:
-    """
-    This abstract class adds the ability to be saveable to inherited classes. The filename can be
-    generated internally by implementing the method get_filename() or provided from outside otherwise.
-    """
-
-    C_SUFFIX = '.pkl'
-
-## -------------------------------------------------------------------------------------------------
-    def generate_filename(self):
-        """
-        To be redefined in case of use of internal generated file names.
-
-        Returns
-        -------
-            Returns an internal unique filename. 
-        """
-
-        raise NotImplementedError
+    def __init__(self, p_id = None):
+        self.set_id(p_id=p_id)
 
 
 ## -------------------------------------------------------------------------------------------------
-    def save(self, p_path, p_filename=None) -> bool:
+    def get_id(self):
+        return self._id
+    
+
+## -------------------------------------------------------------------------------------------------
+    def set_id(self, p_id = None):
         """
-        Saves content to the given path and file name. If file name is None, a unique file name will
-        be generated by calling method generate_filename(). If it returns False then the saving method 
-        is failed. For saving the object, the custom method _save is called. This contains a default
-        implementation based on pickle/dill and can be redefined on demand.
+        Sets/generates a new id.
 
         Parameters
         ----------
-            p_path          Path where file will be saved
-            p_filename      File name (if None an internal filename will be generated)
-
-        Returns
-        -------
-            True, if file content was saved successfully. False otherwise.
+        p_id
+            Optional external id. If None, a unique id is generated.
         """
 
-        if (p_filename is not None) and (p_filename != ''):
-            self.filename = p_filename
+        if p_id is not None:
+            self._id = p_id
         else:
-            self.filename = self.generate_filename()
-
-        if self.filename is None:
-            return False
-
-        if not os.path.exists(p_path):
-            os.makedirs(p_path)
-
-        return self._save( p_path=p_path, p_filename=self.filename)
-
-
-## -------------------------------------------------------------------------------------------------
-    def _save(self, p_path, p_filename) -> bool:
-        """
-        Custom method for saving an object. The default implementation is based on pickle/dill. Redefine
-        on demand. See method save() for further details.
-        """
-
-        pkl.dump( obj=self, 
-                  file=open(p_path + os.sep + p_filename, "wb"),
-                  protocol=pkl.HIGHEST_PROTOCOL )
-        return True
-
-
-
-
-
-## -------------------------------------------------------------------------------------------------
-## -------------------------------------------------------------------------------------------------
-class LoadSave(Loadable, Saveable):
-    """
-    This abstract class adds the ability to be loadable and saveable to inherited classes. The 
-    filename can be generated internally by implementing the method generate_filename() or provided 
-    from outside otherwise. See classes Loadable and Saveable for further information.
-    """
-
-    pass
+            self._id = uuid.uuid4()
 
 
 
@@ -297,6 +239,291 @@ class Log:
 
 
 
+## -------------------------------------------------------------------------------------------------
+## -------------------------------------------------------------------------------------------------
+class Persistent (Id, Log):
+    """
+    Property class that inherits persistence to its child classes.
+
+    Parameters
+    ----------
+    p_id
+        Optional external id
+    p_logging
+        Log level (see constants C_LOG_*). Default: Log.C_LOG_ALL
+
+    Attributes
+    ----------
+    C_PERSISTENCE_VERSION : str
+        Version of the implementation of the persistence. Shall be raised in child classes whenever 
+        an incompatible change has been done.
+    C_SUFFIX : str = '.pkl'
+        Default suffix for pickled result files.
+    """
+
+    C_PERSISTENCE_VERSION : str = '1.0.0'
+    C_SUFFIX : str              = '.pkl'
+
+## -------------------------------------------------------------------------------------------------
+    def __init__(self, p_id=None, p_logging=Log.C_LOG_ALL):
+        Id.__init__(self, p_id=p_id)
+        Log.__init__(self, p_logging=p_logging)
+
+        # Persistent objects need a unique filename
+        self.set_filename( p_filename_stub = self.__class__.__name__ + '[' + str(self.get_id()) + ']')
+        self._persistence_version = self.C_PERSISTENCE_VERSION
+
+
+## -------------------------------------------------------------------------------------------------
+    def get_filename_stub(self) -> str:
+        """
+        Returns the unique filename of the object without a suffix.
+
+        Returns
+        -------
+        filename_stub : str
+            Filename stub.
+        """
+
+        return self._filename_stub
+
+
+## -------------------------------------------------------------------------------------------------
+    def get_filename(self) -> str:
+        """
+        Returns the full unique filename of the object including the suffix.
+
+        Returns
+        -------
+        filename : str
+            Full filename.
+        """
+
+        return self._filename
+    
+
+## -------------------------------------------------------------------------------------------------
+    def set_filename(self, p_filename_stub:str, p_suffix:str=None):
+        self._filename_stub = p_filename_stub
+        
+        if p_suffix is not None:
+            self._suffix = p_suffix 
+        else:
+            self._suffix = self.C_SUFFIX
+
+        self._filename = self._filename_stub + self._suffix
+    
+
+## -------------------------------------------------------------------------------------------------
+    def _get_path(self) -> str:
+        """
+        Internal helper method to determine the current path for loading/saving external data.
+        """
+
+        # 1 Check: is self the path donator?
+        try:
+            return g_persistence_file_paths[self.get_filename()]
+        except:
+            pass
+
+
+        # 2 Locate path donator in call stack
+        frame_id = 1
+        path = ''
+        path_donator_bak = self
+        
+        while True:
+            try:
+                frame = sys._getframe(frame_id)
+            except:
+                break
+
+            try:
+                path_donator = frame.f_locals['self']
+                if path_donator != path_donator_bak:
+                    path_donator_bak = path_donator
+                    filename = path_donator.get_filename()
+                    path = g_persistence_file_paths[filename]
+                    break
+            except:
+                pass
+            
+            frame_id += 1
+
+        return path
+
+
+## -------------------------------------------------------------------------------------------------
+    @classmethod
+    def load( cls, p_path:str, p_filename:str ):
+        """
+        Static method to load an object of the current class from a file using pickle/dill. During
+        unpickling the given file, standard method __setstate__() is called. This in turn is implemented
+        specifically and calls the MLPro custom method _complete_state(). This method allows the
+        completion of the unpickled object from further externally stored data.
+
+        Parameters
+        ----------
+        p_path : str
+            Path where file will be saved
+        p_filename : str = None      
+            File name (if None an internal filename will be used)
+
+        Returns
+        -------
+        Object 
+            Object of the given class that was unpickled from the given file.
+        """
+
+        g_persistence_file_paths[p_filename] = p_path
+
+        obj = pkl.load(open(p_path + os.sep + p_filename, 'rb'))
+
+        obj.log(Log.C_LOG_TYPE_I, 'Object loaded from file "' + p_path + os.sep + p_filename + '"')
+
+        return obj
+    
+
+## -------------------------------------------------------------------------------------------------
+    def __setstate__(self, p_state:dict):
+        """
+        Python standard method to set the internal object state during unpickling from file.
+        The custom method _complete_state() is called to complete the state from further external
+        sources.
+
+        Parameters
+        ----------
+        p_state : dict
+            Incoming object state dictionary to be completed.
+        """
+
+        # 1 Check for compatible version of persistence
+        try:
+            ver_file = p_state['_persistence_version']
+        except:
+            ver_file = None
+
+        if ( ver_file is None ) or ( ver_file != self.C_PERSISTENCE_VERSION ):
+            raise ParamError('Pickle file', p_state['_filename'], 'not compatible!')
+                     
+        # 2 Update object state 
+        self.__dict__.update(p_state)
+
+        # 3 Call custom method to complete the object state
+        self._os_sep = os.sep
+        self._complete_state( p_path = self._get_path(),
+                              p_os_sep = os.sep,
+                              p_filename_stub = self.get_filename_stub() )
+
+
+## -------------------------------------------------------------------------------------------------
+    def _complete_state(self, p_path:str, p_os_sep:str, p_filename_stub:str):
+        """
+        Custom method to complete the object state (=self) from external data sources. This method
+        is called by standard method __setstate__() during unpickling the object from an external
+        file. 
+
+        Parameters
+        ----------
+        p_path : str
+            Path of the object pickle file (and further optional related files)
+        p_os_sep : str
+            OS-specific path separator.
+        p_filename_stub : str
+            Filename stub to be used for further optional custom data files
+        """
+
+        pass
+
+
+## -------------------------------------------------------------------------------------------------
+    def save(self, p_path:str, p_filename:str=None) -> bool:
+        """
+        Saves the object to the given path and file name using pickle/dill. If file name is None, a 
+        unique inernal file name is used (recommended). During pickling the Python standard method
+        __getstate() is called. This in turn is implemented specifically and calls the MLPro custom
+        method _reduce_state(). This method allows to reduce unpickleable components from the object
+        state before pickling. These components can optionally be stored in separate files of a 
+        suitable format.
+
+        Parameters
+        ----------
+        p_path : str
+            Path where file will be saved
+        p_filename : str = None      
+            File name (if None an internal filename will be used)
+
+        Returns
+        -------
+        successful : bool
+            True, if file content was saved successfully. False otherwise.
+        """
+
+        # 1 Create folder if it doesn't exist
+        if not os.path.exists(p_path): os.makedirs(p_path)
+
+        if p_filename is not None:
+            filename = p_filename
+        else:
+            filename = self.get_filename()
+
+        g_persistence_file_paths[filename] = p_path
+
+        pkl.dump( obj=self, 
+                  file=open(p_path + os.sep + filename, "wb"),
+                  protocol=pkl.HIGHEST_PROTOCOL )
+        
+        self.log(Log.C_LOG_TYPE_I, 'Object saved to file "' + p_path + os.sep + filename + '"')
+        return True
+
+
+## -------------------------------------------------------------------------------------------------
+    def __getstate__(self) -> dict:
+        """
+        Python standard method to get the internal object state during pickling from file.
+        The custom method _reduce_state() is called to remove data or object references from the
+        state that can not be pickled.
+
+        Returns
+        ----------
+        state : dict
+            (Reduced) object state dictionary to be pickled.
+        """
+
+        state = self.__dict__.copy()
+
+        self._reduce_state( p_state = state, 
+                            p_path = self._get_path(), 
+                            p_os_sep = os.sep,
+                            p_filename_stub = self.get_filename_stub() )           
+
+        return state
+
+
+## -------------------------------------------------------------------------------------------------
+    def _reduce_state(self, p_state:dict, p_path:str, p_os_sep:str, p_filename_stub:str):
+        """
+        Custom method to reduce the given object state by components that can not be pickled. 
+        Further data files can be created in the given path and should use the given filename stub.
+
+        Parameters
+        ----------
+        p_state : dict
+            Object state dictionary to be reduced by components that can not be pickled.
+        p_path : str
+            Path to store further optional custom data files
+        p_os_sep : str
+            OS-specific path separator.
+        p_filename_stub : str
+            Filename stub to be used for further optional custom data files
+        """
+
+        pass
+
+
+
+
+
 ## -------------------------------------------------------------------------------------------------
 ## -------------------------------------------------------------------------------------------------
 class Timer:
@@ -491,10 +718,11 @@ class  ScientificObject:
 
 
 
+
 ## -------------------------------------------------------------------------------------------------
 ## -------------------------------------------------------------------------------------------------
 
-class PersonalisedStamp:
+class PersonalisedStamp (Id):
     """
     This class serves as a base class of label to set up a name and id for another class.
     
@@ -524,37 +752,6 @@ class PersonalisedStamp:
             raise NotImplementedError('Please add a name!')
         
         self.set_id(p_id)
-            
-
-## -------------------------------------------------------------------------------------------------
-    def set_id(self, p_id:int=None):
-        """
-        This method provides a functionality to set an unique ID.
-
-        Parameters
-        ----------
-        p_id : int, optional
-            An unique ID. Default: None.
-
-        """
-        if p_id is None:
-            self._id = str(uuid.uuid4())
-        else:
-            self._id = str(p_id)
-
-
-## -------------------------------------------------------------------------------------------------
-    def get_id(self) -> str:
-        """
-        This method provides a functionality to get the defined unique ID.
-
-        Returns
-        -------
-        str
-            The unique ID.
-
-        """
-        return self._id
 
 
 ## -------------------------------------------------------------------------------------------------
